/* This file can be redistributed and used under GNU Library General Public
 * License. (c) 2000 Micha Moskal <malekith@topnet.pl>. 
 * $Id: hash.h,v 1.3 2000/09/09 17:59:14 malekith Exp $ */

#ifndef _V_HASH_H
#define _V_HASH_H

#include <string.h>
#include <stdlib.h>

#include <v/utils.h>
#include <v/block.h>

#define DEF_HASH_SIZE 16

// -- helpers --
template <class T>
class HashNode {
public:
	char *key;
	T data;
	HashNode<T> *next;
	
	HashNode(const char *s, const T& ref) 
		: key(strdup(s)), data(ref), next(0) { }
	~HashNode() { free(key); }
	bool operator ==(const char *k)
		{ return strcmp(key, k) == 0; }
};

class HashBase {
protected:
	HashBase() {}
	static int get_hash(const char *key);
};

// -- meat ---
template <class T>
class Hash : public HashBase {
	HashNode<T> **h;
	int nh;
	void init();
public:
	Hash(const T& _undef) { init();	undef = _undef; }
	Hash() { init(); }
	~Hash();
	
	T& get(const char *key);
	void put(const char *key, const T& ref);
	int remove(const char *key);
	T& operator () (const char *key);
	void clear();
	
	const Hash<T> & operator =(const Hash<T> &ref);
	
	Block<T> values();
	Block<String> keys();
	
	T undef;
};

// ------------------------------------------------------------------------
// the implementation
template <class T>
void Hash<T>::init()
{
	h = new HashNode<T>* [nh = DEF_HASH_SIZE];
	for (int i=0; i<nh; i++)
		h[i] = 0;
}

template <class T>
void Hash<T>::clear()
{
	for (int i=0; i<nh; i++) {
		HashNode<T> *p, *t;
		for (p = h[i]; p; p = t) {
			t = p->next;
			delete p;
		}
		h[i] = 0;
	}
}

template <class T>
Hash<T>::~Hash()
{
	for (int i=0; i<nh; i++) {
		HashNode<T> *p, *t;
		for (p = h[i]; p; p = t) {
			t = p->next;
			delete p;
		}
	}
	delete h;
}

template <class T>
void Hash<T>::put(const char *key, const T& ref)
{
	int i = get_hash(key) & (nh - 1);
	HashNode<T> *p = new HashNode<T>(key, ref);
	p->next = h[i];
	h[i] = p;
}

template <class T>
T& Hash<T>::get(const char *key)
{
	HashNode<T> *p;
	
	for (p = h[get_hash(key) & (nh - 1)]; p; p = p->next)
		if (*p == key)
			break;
	if (p)
		return p->data;
	else
		return undef;
}

template <class T>
int Hash<T>::remove(const char *key)
{
	HashNode <T> *p;
	int i = get_hash(key) & (nh - 1);

	if ((p = h[i]) == 0)
		return -1;

	if (*p == key) {
		h[i] = h[i]->next;
		delete p;
		return 0;
	}
	for (; p->next; p = p->next)
		if (*(p->next) == key)
			break;
	if (!p->next)
		return -1;
	p->next = p->next->next;
	delete p->next;
	return 0;
}

template <class T>
T& Hash<T>::operator () (const char *key)
{
	HashNode<T> *p;
	int i = get_hash(key) & (nh - 1);
	
	for (p = h[i]; p; p = p->next)
		if (*p == key)
			break;
	if (p)
		return p->data;
	else {
		HashNode<T> *p = new HashNode<T>(key, undef);
		p->next = h[i];
		h[i] = p;
		return p->data;
	}
}

template <class T>
const Hash<T> & Hash<T>::operator =(const Hash<T> &ref)
{
	this->~Hash();
	init();
	
	for (int i=0; i<ref.nh; i++) {
		const HashNode<T> *p;
		for (p = ref.h[i]; p; p = p->next)
			put(p->key, p->data);
	}
	
	return *this;
}

template <class T>
Block<T> Hash<T>::values()
{
	Block<T> r;
	const HashNode<T> *p;
	
	for (int i=0; i<nh; i++)
		for (p = h[i]; p; p = p->next)
			r.push(p->data);
	return r;
}

template <class T>
Block<String> Hash<T>::keys()
{
	Block<String> r;
	const HashNode<T> *p;
	
	for (int i=0; i<nh; i++)
		for (p = h[i]; p; p = p->next)
			r.push(p->key);
	return r;
}

#endif /* _V_HASH_H */
